from functools import partial

from sklearn.neighbors import KNeighborsTransformer

BallTreeTransformer = partial(KNeighborsTransformer, algorithm="ball_tree")
KDTreeTransformer = partial(KNeighborsTransformer, algorithm="kd_tree")
BruteTransformer = partial(KNeighborsTransformer, algorithm="brute")


__all__ = ["BallTreeTransformer", "BruteTransformer", "KDTreeTransformer"]



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

import os

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
# import sys
# sys.path.insert(0, os.path.abspath('.'))

# -- Project information -----------------------------------------------------

project = "sklearn-ann"
copyright = "2021, Frankie Robertson"
author = "Frankie Robertson"


# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    "sphinx.ext.napoleon",
    "scanpydoc.definition_list_typed_field",
    "scanpydoc.rtd_github_links",
    "sphinx_issues",
    "sphinx.ext.linkcode",
]

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]

autosummary_generate = True
autodoc_default_options = {
    "undoc-members": True,
}


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = "sphinx_book_theme"
html_theme_options = dict(
    repository_url="https://github.com/frankier/sklearn-ann",
    repository_branch=os.environ.get("READTHEDOCS_GIT_IDENTIFIER", "main"),
)
rtd_links_prefix = "src"

autodoc_mock_imports = ["annoy", "faiss", "pynndescent", "nmslib"]



import pytest
from sklearn.utils.estimator_checks import check_estimator

from sklearn_ann.cluster.rnn_dbscan import RnnDBSCAN

ESTIMATORS = [RnnDBSCAN]


@pytest.mark.parametrize("Estimator", ESTIMATORS)
def test_all_estimators(Estimator):
    check_estimator(Estimator())



from sklearn_ann.test_utils import assert_row_close, needs

try:
    from sklearn_ann.kneighbors.faiss import FAISSTransformer
except ImportError:
    pass


@needs.faiss
def test_euclidean(random_small, random_small_pdists):
    trans = FAISSTransformer(metric="euclidean")
    mat = trans.fit_transform(random_small)
    euclidean_dist = random_small_pdists["euclidean"]
    assert_row_close(mat, euclidean_dist)


@needs.faiss
def test_cosine(random_small, random_small_pdists):
    trans = FAISSTransformer(metric="cosine")
    mat = trans.fit_transform(random_small)
    cosine_dist = random_small_pdists["cosine"]
    assert_row_close(mat, cosine_dist)



import numpy as np
from scipy.sparse import csr_matrix
from sklearn.utils.validation import validate_data


def check_metric(metric, metrics):
    if metric not in metrics:
        raise ValueError(f"Unknown metric {metric!r}. Valid metrics are {metrics!r}")


def get_sparse_indices(mat, idx):
    start_idx = mat.indptr[idx]
    end_idx = mat.indptr[idx + 1]
    return mat.indices[start_idx:end_idx]


def get_sparse_row(mat, idx):
    start_idx = mat.indptr[idx]
    end_idx = mat.indptr[idx + 1]
    return zip(mat.indices[start_idx:end_idx], mat.data[start_idx:end_idx])


def trunc_csr(csr, k):
    indptr = np.empty_like(csr.indptr)
    num_rows = len(csr.indptr) - 1
    indices = [None] * num_rows
    data = [None] * num_rows
    cur_indptr = 0
    for row_idx in range(num_rows):
        indptr[row_idx] = cur_indptr
        start_idx = csr.indptr[row_idx]
        old_end_idx = csr.indptr[row_idx + 1]
        end_idx = min(old_end_idx, start_idx + k)
        data[row_idx] = csr.data[start_idx:end_idx]
        indices[row_idx] = csr.indices[start_idx:end_idx]
        ptr_inc = min(k, old_end_idx - start_idx)
        cur_indptr = cur_indptr + ptr_inc
    indptr[-1] = cur_indptr
    return csr_matrix((np.concatenate(data), np.concatenate(indices), indptr))


def or_else_csrs(csr1, csr2):
    # Possible TODO: Could use numba/Cython to speed this up?
    if csr1.shape != csr2.shape:
        raise ValueError("csr1 and csr2 must be the same shape")
    indptr = np.empty_like(csr1.indptr)
    indices = []
    data = []
    for row_idx in range(len(indptr) - 1):
        indptr[row_idx] = len(indices)
        csr1_it = iter(get_sparse_row(csr1, row_idx))
        csr2_it = iter(get_sparse_row(csr2, row_idx))
        cur_csr1 = next(csr1_it, None)
        cur_csr2 = next(csr2_it, None)
        while 1:
            if cur_csr1 is None and cur_csr2 is None:
                break
            elif cur_csr1 is None:
                cur_index, cur_datum = cur_csr2
            elif cur_csr2 is None:
                cur_index, cur_datum = cur_csr1
            elif cur_csr1[0] < cur_csr2[0]:
                cur_index, cur_datum = cur_csr1
                cur_csr1 = next(csr1_it, None)
            elif cur_csr2[0] < cur_csr1[0]:
                cur_index, cur_datum = cur_csr2
                cur_csr2 = next(csr2_it, None)
            else:
                cur_index, cur_datum = cur_csr1
                cur_csr1 = next(csr1_it, None)
                cur_csr2 = next(csr2_it, None)
            indices.append(cur_index)
            data.append(cur_datum)
    indptr[-1] = len(indices)
    return csr_matrix((data, indices, indptr), shape=csr1.shape)


def postprocess_knn_csr(knns, include_fwd=True, include_rev=False):
    if not include_fwd and not include_rev:
        raise ValueError("One of include_fwd or include_rev must be True")
    elif include_rev and not include_fwd:
        return knns.transpose(copy=False)
    elif not include_rev and include_fwd:
        return knns
    else:
        inv_knns = knns.transpose(copy=True)
        return or_else_csrs(knns, inv_knns)


class TransformerChecksMixin:
    def _transform_checks(self, X, *fitted_props, **check_params):
        from sklearn.utils.validation import check_is_fitted

        X = validate_data(self, X, reset=False, **check_params)
        check_is_fitted(self, *fitted_props)
        return X



"""
===================================
Demo of RNN-DBSCAN clustering algorithm
===================================

Finds core samples of high density and expands clusters from them.

Mostly copypasted from sklearn's DBSCAN example.

"""

import numpy as np
from sklearn import metrics
from sklearn.datasets import make_blobs
from sklearn.preprocessing import StandardScaler

from sklearn_ann.cluster.rnn_dbscan import RnnDBSCAN

# #############################################################################
# Generate sample data
centers = [[1, 1], [-1, -1], [1, -1]]
X, labels_true = make_blobs(
    n_samples=750, centers=centers, cluster_std=0.4, random_state=0
)

X = StandardScaler().fit_transform(X)

# #############################################################################
# Compute DBSCAN
db = RnnDBSCAN(n_neighbors=10).fit(X)
core_samples_mask = np.zeros_like(db.labels_, dtype=bool)
core_samples_mask[db.core_sample_indices_] = True
labels = db.labels_

# Number of clusters in labels, ignoring noise if present.
n_clusters_ = len(set(labels)) - (1 if -1 in labels else 0)
n_noise_ = list(labels).count(-1)

print(f"""\
Estimated number of clusters: {n_clusters_}
Estimated number of noise points: {n_noise_}
Homogeneity: {metrics.homogeneity_score(labels_true, labels):0.3f}
Completeness: {metrics.completeness_score(labels_true, labels):0.3f}
V-measure: {metrics.v_measure_score(labels_true, labels):0.3f}
Adjusted Rand Index: {metrics.adjusted_rand_score(labels_true, labels):0.3f}
Adjusted Mutual Info: {metrics.adjusted_mutual_info_score(labels_true, labels):0.3f}
Silhouette Coefficient: {metrics.silhouette_score(X, labels):0.3f}\
""")

# #############################################################################
# Plot result
import matplotlib.pyplot as plt

# Black removed and is used for noise instead.
unique_labels = set(labels)
colors = [plt.cm.Spectral(each) for each in np.linspace(0, 1, len(unique_labels))]
for k, col in zip(unique_labels, colors):
    if k == -1:
        # Black used for noise.
        col = [0, 0, 0, 1]

    class_member_mask = labels == k

    xy = X[class_member_mask & core_samples_mask]
    plt.plot(
        xy[:, 0],
        xy[:, 1],
        "o",
        markerfacecolor=tuple(col),
        markeredgecolor="k",
        markersize=14,
    )

    xy = X[class_member_mask & ~core_samples_mask]
    plt.plot(
        xy[:, 0],
        xy[:, 1],
        "o",
        markerfacecolor=tuple(col),
        markeredgecolor="k",
        markersize=6,
    )

plt.title(f"Estimated number of clusters: {n_clusters_}")
plt.show()



from collections import deque
from typing import cast

import numpy as np
from sklearn.base import BaseEstimator, ClusterMixin
from sklearn.neighbors import KNeighborsTransformer
from sklearn.utils import Tags
from sklearn.utils.validation import validate_data

from ..utils import get_sparse_row

UNCLASSIFIED = -2
NOISE = -1


def join(it1, it2):
    cur_it1 = next(it1, None)
    cur_it2 = next(it2, None)
    while 1:
        if cur_it1 is None and cur_it2 is None:
            break
        elif cur_it1 is None:
            yield cur_it2
            cur_it2 = next(it2, None)
        elif cur_it2 is None:
            yield cur_it1
            cur_it1 = next(it1, None)
        elif cur_it1[0] == cur_it2[0]:
            yield cur_it1
            cur_it1 = next(it1, None)
            cur_it2 = next(it2, None)
        elif cur_it1[0] < cur_it2[0]:
            yield cur_it1
            cur_it1 = next(it1, None)
        else:
            yield cur_it2
            cur_it2 = next(it2, None)


def neighborhood(is_core, knns, rev_knns, idx):
    # TODO: Make this inner bit faster
    knn_it = get_sparse_row(knns, idx)
    rev_core_knn_it = (
        (other_idx, dist)
        for other_idx, dist in get_sparse_row(rev_knns, idx)
        if is_core[other_idx]
    )
    yield from (
        (other_idx, dist)
        for other_idx, dist in join(knn_it, rev_core_knn_it)
        if other_idx != idx
    )


def rnn_dbscan_inner(is_core, knns, rev_knns, labels):
    cluster = 0
    cur_dens = 0
    dens = []
    for x_idx in range(len(labels)):
        if labels[x_idx] == UNCLASSIFIED:
            # Expand cluster
            if is_core[x_idx]:
                labels[x_idx] = cluster
                # TODO: Make this inner bit faster - can just assume
                # sorted an keep sorted
                seeds = deque()
                for neighbor_idx, dist in neighborhood(is_core, knns, rev_knns, x_idx):
                    labels[neighbor_idx] = cluster
                    if dist > cur_dens:
                        cur_dens = dist
                    seeds.append(neighbor_idx)
                while seeds:
                    y_idx = seeds.popleft()
                    if is_core[y_idx]:
                        for z_idx, dist in neighborhood(is_core, knns, rev_knns, y_idx):
                            if dist > cur_dens:
                                cur_dens = dist
                            if labels[z_idx] == UNCLASSIFIED:
                                seeds.append(z_idx)
                                labels[z_idx] = cluster
                            elif labels[z_idx] == NOISE:
                                labels[z_idx] = cluster
                dens.append(cur_dens)
                cur_dens = 0
                cluster += 1
            else:
                labels[x_idx] = NOISE
    # Expand clusters
    for x_idx in range(len(labels)):
        if labels[x_idx] == NOISE:
            min_cluster = NOISE
            min_dist = float("inf")
            for n_idx, n_dist in get_sparse_row(knns, x_idx):
                if n_dist >= min_dist or not is_core[n_idx]:
                    continue
                cluster = labels[n_idx]
                if n_dist > dens[cluster]:
                    continue
                min_cluster = cluster
                min_dist = n_dist
            labels[x_idx] = min_cluster
    return dens


class RnnDBSCAN(ClusterMixin, BaseEstimator):
    """
    Implements the RNN-DBSCAN clustering algorithm.

    Parameters
    ----------
    n_neighbors : int
        The number of neighbors in the kNN-graph (the k in kNN), and the
        theshold of reverse nearest neighbors for a node to be considered a
        core node.
    input_guarantee : "none" | "kneighbors"
        A guarantee on input matrices. If equal to "kneighbors", the algorithm
        will assume you are passing in the kNN graph exactly as required, e.g.
        with n_neighbors. This can be used to pass in a graph produced by one
        of the implementations of the KNeighborsTransformer interface.
    n_jobs : int
        The number of jobs to use. Currently has not effect since no part of
        the algorithm has been parallelled.
    keep_knns : bool
        If true, the kNN and inverse kNN graph will be saved to `knns_` and
        `rev_knns_` after fitting.

    See Also
    --------
    simple_rnn_dbscan_pipeline:
        Create a pipeline of a KNeighborsTransformer and RnnDBSCAN

    References
    ----------
    A. Bryant and K. Cios, "RNN-DBSCAN: A Density-Based Clustering
    Algorithm Using Reverse Nearest Neighbor Density Estimates," in IEEE
    Transactions on Knowledge and Data Engineering, vol. 30, no. 6, pp.
    1109-1121, 1 June 2018, doi: 10.1109/TKDE.2017.2787640.
    """

    def __init__(
        self, n_neighbors=5, *, input_guarantee="none", n_jobs=None, keep_knns=False
    ):
        self.n_neighbors = n_neighbors
        self.input_guarantee = input_guarantee
        self.n_jobs = n_jobs
        self.keep_knns = keep_knns

    def fit(self, X, y=None):
        X = validate_data(self, X, accept_sparse="csr")
        if self.input_guarantee == "none":
            algorithm = KNeighborsTransformer(n_neighbors=self.n_neighbors)
            X = algorithm.fit_transform(X)
        elif self.input_guarantee == "kneighbors":
            pass
        else:
            raise ValueError(
                "Expected input_guarantee to be one of 'none', 'kneighbors'"
            )

        XT = X.transpose().tocsr(copy=True)
        if self.keep_knns:
            self.knns_ = X
            self.rev_knns_ = XT

        # Initially, all samples are unclassified.
        labels = np.full(X.shape[0], UNCLASSIFIED, dtype=np.int32)

        # A list of all core samples found. -1 is to account for diagonal.
        core_samples = XT.getnnz(1) - 1 >= self.n_neighbors

        dens = rnn_dbscan_inner(core_samples, X, XT, labels)

        self.core_sample_indices_ = core_samples.nonzero()
        self.labels_ = labels
        self.dens_ = dens

        return self

    def fit_predict(self, X, y=None):
        self.fit(X, y=y)
        return self.labels_

    def drop_knns(self):
        del self.knns_
        del self.rev_knns_

    def __sklearn_tags__(self) -> Tags:
        tags = cast(Tags, super().__sklearn_tags__())
        tags.input_tags.sparse = True
        return tags


def simple_rnn_dbscan_pipeline(
    neighbor_transformer, n_neighbors, n_jobs=None, keep_knns=None, **kwargs
):
    """
    Create a simple pipeline comprising a transformer and RnnDBSCAN.

    Parameters
    ----------
    neighbor_transformer : class implementing KNeighborsTransformer interface
    n_neighbors:
        Passed to neighbor_transformer and RnnDBSCAN
    n_jobs:
        Passed to neighbor_transformer and RnnDBSCAN
    keep_knns:
        Passed to RnnDBSCAN
    kwargs:
        Passed to neighbor_transformer
    """
    from sklearn.pipeline import make_pipeline

    return make_pipeline(
        neighbor_transformer(n_neighbors=n_neighbors, n_jobs=n_jobs, **kwargs),
        RnnDBSCAN(
            n_neighbors=n_neighbors,
            input_guarantee="kneighbors",
            n_jobs=n_jobs,
            keep_knns=keep_knns,
        ),
    )



"""
=======================================================
Demo of RnnDBSCAN clustering algorithm on large dataset
=======================================================

Tests RnnDBSCAN on a large dataset. Requires pandas.

"""

import numpy as np
from joblib import Memory
from sklearn import metrics
from sklearn.datasets import fetch_openml

from sklearn_ann.cluster.rnn_dbscan import simple_rnn_dbscan_pipeline


# #############################################################################
# Generate sample data
def fetch_mnist():
    print("Downloading mnist_784")
    mnist = fetch_openml("mnist_784")
    return mnist.data / 255, mnist.target


memory = Memory("./mnist")

X, y = memory.cache(fetch_mnist)()


def run_rnn_dbscan(neighbor_transformer, n_neighbors, **kwargs):
    # #############################################################################
    # Compute RnnDBSCAN

    pipeline = simple_rnn_dbscan_pipeline(neighbor_transformer, n_neighbors, **kwargs)
    labels = pipeline.fit_predict(X)
    db = pipeline.named_steps["rnndbscan"]
    core_samples_mask = np.zeros_like(db.labels_, dtype=bool)
    core_samples_mask[db.core_sample_indices_] = True

    # Number of clusters in labels, ignoring noise if present.
    n_clusters_ = len(set(labels)) - (1 if -1 in labels else 0)
    n_noise_ = list(labels).count(-1)

    print(f"""\
Estimated number of clusters: {n_clusters_}
Estimated number of noise points: {n_noise_}
Homogeneity: {metrics.homogeneity_score(y, labels):0.3f}
Completeness: {metrics.completeness_score(y, labels):0.3f}
V-measure: {metrics.v_measure_score(y, labels):0.3f}
Adjusted Rand Index: {metrics.adjusted_rand_score(y, labels):0.3f}
Adjusted Mutual Information: {metrics.adjusted_mutual_info_score(y, labels):0.3f}
Silhouette Coefficient: {metrics.silhouette_score(X, labels):0.3f}\
""")


if __name__ == "__main__":
    import code

    print("""\
Now you can import your chosen transformer_cls and run:
run_rnn_dbscan(transformer_cls, n_neighbors, **params)
e.g.
from sklearn_ann.kneighbors.pynndescent import PyNNDescentTransformer
run_rnn_dbscan(PyNNDescentTransformer, 10)\
""")
    code.interact(local=locals())



from __future__ import annotations

import math

import faiss
import numpy as np
from faiss import normalize_L2
from joblib import cpu_count
from scipy.sparse import csr_matrix
from sklearn.base import BaseEstimator, TransformerMixin
from sklearn.utils import Tags, TargetTags, TransformerTags
from sklearn.utils.validation import validate_data

from ..utils import TransformerChecksMixin, postprocess_knn_csr

L2_INFO = {"metric": faiss.METRIC_L2, "sqrt": True}


METRIC_MAP = {
    "cosine": {
        "metric": faiss.METRIC_INNER_PRODUCT,
        "normalize": True,
        "negate": True,
    },
    "l1": {"metric": faiss.METRIC_L1},
    "cityblock": {"metric": faiss.METRIC_L1},
    "manhattan": {"metric": faiss.METRIC_L1},
    "l2": L2_INFO,
    "euclidean": L2_INFO,
    "sqeuclidean": {"metric": faiss.METRIC_L2},
    "canberra": {"metric": faiss.METRIC_Canberra},
    "braycurtis": {"metric": faiss.METRIC_BrayCurtis},
    "jensenshannon": {"metric": faiss.METRIC_JensenShannon},
}


def mk_faiss_index(feats, inner_metric, index_key="", nprobe=128) -> faiss.Index:
    size, dim = feats.shape
    if not index_key:
        if inner_metric == faiss.METRIC_INNER_PRODUCT:
            index = faiss.IndexFlatIP(dim)
        else:
            index = faiss.IndexFlatL2(dim)
    else:
        if index_key.find("HNSW") < 0:
            raise NotImplementedError(
                "HNSW not implemented: returns distances insted of sims"
            )
        nlist = min(4096, 8 * round(math.sqrt(size)))
        if index_key == "IVF":
            quantizer = index
            index = faiss.IndexIVFFlat(quantizer, dim, nlist, inner_metric)
        else:
            index = faiss.index_factory(dim, index_key, inner_metric)
        if index_key.find("Flat") < 0:
            assert not index.is_trained
        index.train(feats)
        index.nprobe = min(nprobe, nlist)
        assert index.is_trained
    index.add(feats)
    return index


class FAISSTransformer(TransformerChecksMixin, TransformerMixin, BaseEstimator):
    def __init__(
        self,
        n_neighbors=5,
        *,
        metric="euclidean",
        index_key="",
        n_probe=128,
        n_jobs=-1,
        include_fwd=True,
        include_rev=False,
    ):
        self.n_neighbors = n_neighbors
        self.metric = metric
        self.index_key = index_key
        self.n_probe = n_probe
        self.n_jobs = n_jobs
        self.include_fwd = include_fwd
        self.include_rev = include_rev

    @property
    def _metric_info(self):
        return METRIC_MAP[self.metric]

    def fit(self, X, y=None):
        normalize = self._metric_info.get("normalize", False)
        X = validate_data(self, X, dtype=np.float32, copy=normalize)
        self.n_samples_fit_ = X.shape[0]
        if self.n_jobs == -1:
            n_jobs = cpu_count()
        else:
            n_jobs = self.n_jobs
        faiss.omp_set_num_threads(n_jobs)
        inner_metric = self._metric_info["metric"]
        if normalize:
            normalize_L2(X)
        self.faiss_ = mk_faiss_index(X, inner_metric, self.index_key, self.n_probe)
        return self

    def transform(self, X):
        normalize = self._metric_info.get("normalize", False)
        X = self._transform_checks(X, "faiss_", dtype=np.float32, copy=normalize)
        if normalize:
            normalize_L2(X)
        return self._transform(X)

    def _transform(self, X):
        n_samples_transform = self.n_samples_fit_ if X is None else X.shape[0]
        n_neighbors = self.n_neighbors + 1
        if X is None:
            sims, nbrs = self.faiss_.search(
                np.reshape(
                    faiss.rev_swig_ptr(
                        self.faiss_.get_xb(), self.faiss_.ntotal * self.faiss_.d
                    ),
                    (self.faiss_.ntotal, self.faiss_.d),
                ),
                k=n_neighbors,
            )
        else:
            sims, nbrs = self.faiss_.search(X, k=n_neighbors)
        dist_arr = np.array(sims, dtype=np.float32)
        if self._metric_info.get("sqrt", False):
            dist_arr = np.sqrt(dist_arr)
        if self._metric_info.get("negate", False):
            dist_arr = 1 - dist_arr
        del sims
        nbr_arr = np.array(nbrs, dtype=np.int32)
        del nbrs
        indptr = np.arange(0, n_samples_transform * n_neighbors + 1, n_neighbors)
        """
        dist_arr = np.concatenate(
            [
                np.zeros(
                    (n_samples_transform, 1),
                    dtype=dist_arr.dtype
                ),
                dist_arr
            ], axis=1
        )
        nbr_arr = np.concatenate(
            [
                np.arange(n_samples_transform)[:, np.newaxis],
                nbr_arr
            ], axis=1
        )
        """
        mat = csr_matrix(
            (dist_arr.ravel(), nbr_arr.ravel(), indptr),
            shape=(n_samples_transform, self.n_samples_fit_),
        )
        return postprocess_knn_csr(
            mat, include_fwd=self.include_fwd, include_rev=self.include_rev
        )

    def fit_transform(self, X, y=None):
        return self.fit(X, y=y)._transform(X=None)

    def __sklearn_tags__(self) -> Tags:
        return Tags(
            estimator_type="transformer",
            target_tags=TargetTags(required=False),
            transformer_tags=TransformerTags(preserves_dtype=[np.float32]),
            # Could be made deterministic *if* we could reset FAISS's internal RNG
            non_deterministic=True,
        )



import numpy as np
import pytest

from sklearn_ann.test_utils import assert_row_close, needs

try:
    from sklearn_ann.kneighbors.annoy import AnnoyTransformer
except ImportError:
    pass


@needs.annoy
def test_euclidean(random_small, random_small_pdists):
    trans = AnnoyTransformer(metric="euclidean")
    mat = trans.fit_transform(random_small)
    euclidean_dist = random_small_pdists["euclidean"]
    assert_row_close(mat, euclidean_dist)


@needs.annoy
@pytest.mark.xfail(reason="not sure why this isn't working")
def test_angular(random_small, random_small_pdists):
    trans = AnnoyTransformer(metric="angular")
    mat = trans.fit_transform(random_small)
    angular_dist = np.arccos(1 - random_small_pdists["cosine"])
    assert_row_close(mat, angular_dist)



import pytest
from numpy.random import default_rng
from scipy.spatial.distance import pdist, squareform


@pytest.fixture(scope="module")
def random_small():
    gen = default_rng(42)
    return 2 * gen.random((64, 128)) - 1


@pytest.fixture(scope="module")
def random_small_pdists(random_small):
    return {
        metric: squareform(pdist(random_small, metric=metric))
        for metric in ["euclidean", "cosine"]
    }



from pynndescent import PyNNDescentTransformer as PyNNDescentTransformerBase


def no_op():
    pass


class PyNNDescentTransformer(PyNNDescentTransformerBase):
    def fit(self, X, compress_index=True):
        super().fit(X, compress_index=compress_index)
        self.index_.compress_index = no_op
        return self


__all__ = ["PyNNDescentTransformer"]



from enum import Enum
from importlib.util import find_spec


def assert_row_close(sp_mat, actual_pdist, row=42, thresh=0.01):
    row_mat = sp_mat.getrow(row)
    for col, val in zip(row_mat.indices, row_mat.data):
        assert abs(actual_pdist[row, col] - val) < thresh


class needs(Enum):
    """
    A pytest mark generator for skipping tests if a package is not installed.

    Can be used as a decorator:

    >>> @needs.faiss
    >>> def test_x(): pass

    or be called to create a mark object:

    >>> pytest.param(..., marks=[needs.annoy()])
    """

    annoy = ("annoy",)
    faiss = ("faiss-cpu", "faiss-gpu")
    nmslib = ("nmslib",)
    pynndescent = ("pynndescent",)

    def __call__(self, fn=None):
        import pytest

        what = (
            f"package {self.value[0]}"
            if len(self.value) == 1
            else f"one of the packages {set(self.value)}"
        )
        mark = pytest.mark.skipif(
            not find_spec(self.name),
            reason=f"`import {self.name}` needs {what} installed.",
        )
        return mark if fn is None else mark(fn)



import annoy
import numpy as np
from scipy.sparse import csr_matrix
from sklearn.base import BaseEstimator, TransformerMixin
from sklearn.utils import Tags, TargetTags, TransformerTags
from sklearn.utils.validation import validate_data

from ..utils import TransformerChecksMixin


class AnnoyTransformer(TransformerChecksMixin, TransformerMixin, BaseEstimator):
    """Wrapper for using annoy.AnnoyIndex as sklearn's KNeighborsTransformer"""

    def __init__(self, n_neighbors=5, *, metric="euclidean", n_trees=10, search_k=-1):
        self.n_neighbors = n_neighbors
        self.n_trees = n_trees
        self.search_k = search_k
        self.metric = metric

    def fit(self, X, y=None):
        X = validate_data(self, X)
        self.n_samples_fit_ = X.shape[0]
        metric = self.metric if self.metric != "sqeuclidean" else "euclidean"
        self.annoy_ = annoy.AnnoyIndex(X.shape[1], metric=metric)
        for i, x in enumerate(X):
            self.annoy_.add_item(i, x.tolist())
        self.annoy_.build(self.n_trees)
        return self

    def transform(self, X):
        X = self._transform_checks(X, "annoy_")
        return self._transform(X)

    def fit_transform(self, X, y=None):
        return self.fit(X)._transform(X=None)

    def _transform(self, X):
        """As `transform`, but handles X is None for faster `fit_transform`."""

        n_samples_transform = self.n_samples_fit_ if X is None else X.shape[0]

        # For compatibility reasons, as each sample is considered as its own
        # neighbor, one extra neighbor will be computed.
        n_neighbors = self.n_neighbors + 1

        indices = np.empty((n_samples_transform, n_neighbors), dtype=int)
        distances = np.empty((n_samples_transform, n_neighbors))

        if X is None:
            for i in range(self.annoy_.get_n_items()):
                ind, dist = self.annoy_.get_nns_by_item(
                    i, n_neighbors, self.search_k, include_distances=True
                )

                indices[i], distances[i] = ind, dist
        else:
            for i, x in enumerate(X):
                indices[i], distances[i] = self.annoy_.get_nns_by_vector(
                    x.tolist(), n_neighbors, self.search_k, include_distances=True
                )

        if self.metric == "sqeuclidean":
            distances **= 2

        indptr = np.arange(0, n_samples_transform * n_neighbors + 1, n_neighbors)
        kneighbors_graph = csr_matrix(
            (distances.ravel(), indices.ravel(), indptr),
            shape=(n_samples_transform, self.n_samples_fit_),
        )

        return kneighbors_graph

    def __sklearn_tags__(self) -> Tags:
        return Tags(
            estimator_type="transformer",
            target_tags=TargetTags(required=False),
            transformer_tags=TransformerTags(),
        )



import nmslib
import numpy as np
from scipy.sparse import csr_matrix
from sklearn.base import BaseEstimator, TransformerMixin
from sklearn.utils import Tags, TargetTags, TransformerTags
from sklearn.utils.validation import validate_data

from ..utils import TransformerChecksMixin, check_metric

# see more metric in the manual
# https://github.com/nmslib/nmslib/tree/master/manual
METRIC_MAP = {
    "sqeuclidean": "l2",
    "euclidean": "l2",
    "cosine": "cosinesimil",
    "l1": "l1",
    "l2": "l2",
}


class NMSlibTransformer(TransformerChecksMixin, TransformerMixin, BaseEstimator):
    """Wrapper for using nmslib as sklearn's KNeighborsTransformer"""

    def __init__(
        self, n_neighbors=5, *, metric="euclidean", method="sw-graph", n_jobs=1
    ):
        self.n_neighbors = n_neighbors
        self.method = method
        self.metric = metric
        self.n_jobs = n_jobs

    def fit(self, X, y=None):
        X = validate_data(self, X)
        self.n_samples_fit_ = X.shape[0]

        check_metric(self.metric, METRIC_MAP)
        space = METRIC_MAP[self.metric]

        self.nmslib_ = nmslib.init(method=self.method, space=space)
        self.nmslib_.addDataPointBatch(X)
        self.nmslib_.createIndex()
        return self

    def transform(self, X):
        X = self._transform_checks(X, "nmslib_")
        n_samples_transform = X.shape[0]

        # For compatibility reasons, as each sample is considered as its own
        # neighbor, one extra neighbor will be computed.
        n_neighbors = self.n_neighbors + 1

        results = self.nmslib_.knnQueryBatch(X, k=n_neighbors, num_threads=self.n_jobs)
        indices, distances = zip(*results)
        indices, distances = np.vstack(indices), np.vstack(distances)

        if self.metric == "sqeuclidean":
            distances **= 2

        indptr = np.arange(0, n_samples_transform * n_neighbors + 1, n_neighbors)
        kneighbors_graph = csr_matrix(
            (distances.ravel(), indices.ravel(), indptr),
            shape=(n_samples_transform, self.n_samples_fit_),
        )

        return kneighbors_graph

    def __sklearn_tags__(self) -> Tags:
        return Tags(
            estimator_type="transformer",
            target_tags=TargetTags(required=False),
            transformer_tags=TransformerTags(preserves_dtype=[np.float32]),
        )



import numpy as np
import pytest
from sklearn.utils.estimator_checks import check_estimator

from sklearn_ann.test_utils import needs

try:
    from sklearn_ann.kneighbors.annoy import AnnoyTransformer
except ImportError:
    AnnoyTransformer = "AnnoyTransformer"
try:
    from sklearn_ann.kneighbors.faiss import FAISSTransformer
except ImportError:
    FAISSTransformer = "FAISSTransformer"
try:
    from sklearn_ann.kneighbors.nmslib import NMSlibTransformer
except ImportError:
    NMSlibTransformer = "NMSlibTransformer"
try:
    from sklearn_ann.kneighbors.pynndescent import PyNNDescentTransformer
except ImportError:
    PyNNDescentTransformer = "PyNNDescentTransformer"
from sklearn_ann.kneighbors.sklearn import BallTreeTransformer, KDTreeTransformer

ESTIMATORS = [
    pytest.param(AnnoyTransformer, marks=[needs.annoy()]),
    pytest.param(FAISSTransformer, marks=[needs.faiss()]),
    pytest.param(NMSlibTransformer, marks=[needs.nmslib()]),
    pytest.param(PyNNDescentTransformer, marks=[needs.pynndescent()]),
    pytest.param(BallTreeTransformer),
    pytest.param(KDTreeTransformer),
]

PER_ESTIMATOR_XFAIL_CHECKS = {
    AnnoyTransformer: dict(check_estimators_pickle="Cannot pickle AnnoyIndex"),
    FAISSTransformer: dict(
        check_estimators_pickle="Cannot pickle FAISS index",
        check_methods_subset_invariance="Unable to reset FAISS internal RNG",
    ),
    NMSlibTransformer: dict(check_estimators_pickle="Cannot pickle NMSLib index"),
}


def add_mark(param, mark):
    return pytest.param(*param.values, marks=[*param.marks, mark], id=param.id)


@pytest.mark.parametrize(
    "Estimator",
    [
        add_mark(
            est,
            pytest.mark.xfail(
                reason="cannot deal with all dtypes (problem is upsteam)"
            ),
        )
        if est.values[0] is PyNNDescentTransformer
        else est
        for est in ESTIMATORS
    ],
)
def test_all_estimators(Estimator):
    check_estimator(
        Estimator(),
        expected_failed_checks=PER_ESTIMATOR_XFAIL_CHECKS.get(Estimator, {}),
    )


# The following critera are from:
#   https://scikit-learn.org/stable/modules/neighbors.html#nearest-neighbors-transformer
# * only explicitly store nearest neighborhoods of each sample with respect to the
#   training data. This should include those at 0 distance from a query point,
#   including the matrix diagonal when computing the nearest neighborhoods between the
#   training data and itself.
# * each row’s data should store the distance in increasing order
#   (optional. Unsorted data will be stable-sorted, adding a computational overhead).
# * all values in data should be non-negative.
# * there should be no duplicate indices in any row (see https://github.com/scipy/scipy/issues/5807).
# * if the algorithm being passed the precomputed matrix uses k nearest neighbors
#   (as opposed to radius neighborhood), at least k neighbors must be stored in each row
#   (or k+1, as explained in the following note).


def mark_diagonal_0_xfail(est):
    """Mark flaky tests as xfail(strict=False)."""
    # Should probably postprocess these...
    reasons = {
        PyNNDescentTransformer: "sometimes doesn't return diagonal==0",
        FAISSTransformer: "sometimes returns diagonal==eps where eps is small",
    }
    [val] = est.values
    name = val.__name__ if isinstance(val, type) else val
    if reason := reasons.get(val):
        return add_mark(est, pytest.mark.xfail(reason=f"{name} {reason}", strict=False))
    return est


@pytest.mark.parametrize(
    "Estimator", [mark_diagonal_0_xfail(est) for est in ESTIMATORS]
)
def test_all_return_diagonal_0(random_small, Estimator):
    # * only explicitly store nearest neighborhoods of each sample with respect to the
    #   training data. This should include those at 0 distance from a query point,
    #   including the matrix diagonal when computing the nearest neighborhoods
    #   between the training data and itself.

    # Check: do we alway get an "extra" neighbour (diagonal/self)
    est = Estimator(n_neighbors=3)
    knns = est.fit_transform(random_small)
    assert (knns.getnnz(1) == 4).all()

    # Check: diagonal is 0
    next_expected_diagonal = 0
    for row_idx in range(knns.shape[0]):
        start_idx = knns.indptr[row_idx]
        end_idx = knns.indptr[row_idx + 1]
        for col_idx, val in zip(
            knns.indices[start_idx:end_idx], knns.data[start_idx:end_idx]
        ):
            print("self0", row_idx, start_idx, end_idx, col_idx, val)
            if row_idx != col_idx:
                continue
            assert col_idx == next_expected_diagonal
            assert val == 0
            next_expected_diagonal += 1
    assert next_expected_diagonal == len(random_small)


@pytest.mark.parametrize("Estimator", ESTIMATORS)
def test_all_same(random_small, Estimator):
    # Again but for the case of the same element
    ones = np.ones((64, 4))
    est = Estimator(n_neighbors=3)
    knns = est.fit_transform(ones)
    print("knns", knns)
    assert (knns.getnnz(1) == 4).all()
    assert len(knns.nonzero()[0]) == 0
